unit main;

interface

uses
  Windows, Dialogs, Classes, Controls, Menus, SysUtils,
  ToolsAPI, TypInfo,
  Clipbrd,
  CnWizMenuHook,
  HelperTools,
  CopySetting;

type
  TWizardMod = class(TInterfacedObject, IOTAWizard, IOTANotifier)
  protected
    CPUWin : TWinControl;
    FOldExecute : TNotifyEvent;
    procedure Copy50Line(Sender : TObject);
    procedure CopyLine(Sender : TObject);
    procedure InstallExpert(Sender : TObject);
    function Call_Disassemble(Line : Integer; StartType : DWORD) : String;
  public
    procedure AfterSave;           // IOTANotifier
    procedure BeforeSave;
    procedure Modified;
    procedure Destroyed;

    function GetIDString: string;  // IOTAWizard
    function GetName: string;
    function GetState: TWizardState;
    procedure Execute;

    constructor Create;
  end;

implementation

type
  TOnDisassemble = procedure (Sender: TObject; var Address: Integer; var Result: String; var InstSize: Integer) of object;

{ RTTI }
{ Quick Access RTTI Info }

function RTTI_GetMethodAddr(AObject : TObject; propName : String) : TMethod;
begin
{$IFDEF VER120}  //for D4 only
  Result := GetMethodProp(AObject, GetPropInfo(PTypeInfo(AObject.ClassType.ClassInfo), propName));
{$ELSE}
  Result := GetMethodProp(AObject, GetPropInfo(AObject, propName));
{$ENDIF}
end;

function RTTI_GetIntegerProp(AObject : TObject; propName : String) : Integer;
begin
{$IFDEF VER120}  //for D4 only
  Result := GetOrdProp(AObject, GetPropInfo(PTypeInfo(AObject.ClassType.ClassInfo), propName));
{$ELSE}
  Result := GetOrdProp(AObject, GetPropInfo(AObject, propName));
{$ENDIF}
end;

function RTTI_GetStringProp(AObject : TObject; propName : String) : String;
begin
{$IFDEF VER120}  //for D4 only
  Result := GetStrProp(AObject, GetPropInfo(PTypeInfo(AObject.ClassType.ClassInfo), propName));
{$ELSE}
  Result := GetStrProp(AObject, GetPropInfo(AObject, propName));
{$ENDIF}
end;

procedure RTTI_SetMethodAddr(AObject : TObject; propName : String; const Value: TMethod);
begin
{$IFDEF VER120}  //for D4 only
  SetMethodProp(AObject, GetPropInfo(PTypeInfo(AObject.ClassType.ClassInfo), propName), Value);
{$ELSE}
  SetMethodProp(AObject, GetPropInfo(AObject, propName), Value);
{$ENDIF}
end;

procedure RTTI_SetIntegerProp(AObject : TObject; propName : String; const Value: Integer);
begin
{$IFDEF VER120}  //for D4 only
  SetOrdProp(AObject, GetPropInfo(PTypeInfo(AObject.ClassType.ClassInfo), propName), Value);
{$ELSE}
  SetOrdProp(AObject, GetPropInfo(AObject, propName), Value);
{$ENDIF}
end;

{ TWizardMod }
{ A mini Expert Mod. }

function TWizardMod.GetIDString: string;
begin
  Result := 'Aimingoo.CpuWHelper';
end;

function TWizardMod.GetName: string;
begin
  Result := 'CpuWHelper WizardMod';
end;

procedure TWizardMod.AfterSave;
begin
end;

procedure TWizardMod.BeforeSave;
begin
end;

procedure TWizardMod.Destroyed;
begin
end;

procedure TWizardMod.Modified;
begin
end;

procedure TWizardMod.Execute;
begin
end;

function TWizardMod.GetState: TWizardState;
begin
  Result := [];
end;

{ some menu event at here }

procedure TWizardMod.Copy50Line(Sender : TObject);
var
  Code : String;
begin
  if SettingToAll then
    Code := Call_Disassemble(CopyLineCount, CopyLineMode)
  else
    Code := Call_Disassemble(30, stFromTopAddr);

  Clipboard.SetTextBuf(pChar(Code));
end;

procedure TWizardMod.CopyLine(Sender : TObject);
var
  Code : String;
begin
  if TSettingForm.Create(CPUWin).ShowModal = mrYes then
  begin
    Code := Call_Disassemble(CopyLineCount, CopyLineMode);
    Clipboard.SetTextBuf(pChar(Code));
  end;
end;

{ Kernel Code from here }

var
  MainForm        : TWinControl;
  DebugCPUCommand : TObject;

const
  Str_TDisassemblerViewClass = 'TDisassemblerView';
  Str_DisassemblyViewClass = 'TDisassemblyView';
  Str_Delphi32AppClass = 'TAppBuilder';
  Str_DisasEvent = 'OnDisassemble';
  Str_ExecuteCpuWin = 'OnExecute';
  Str_CPUPaneMenu = 'CPUPaneMenu';
  Str_CpuCommandAction = 'DebugCPUCommand';

function TWizardMod.Call_Disassemble(Line : Integer; StartType : DWORD) : String;
var
  S_ASM : String;
  S_Source : String;

  Old_P : Integer;
  Old_TopP : Integer;

  FDisassemble : TOnDisassemble;
  WNDs     : TObject;
  aHwnd    : HWND;
  DisComp  : TWinControl;

  P : Integer;
  L : Integer;
begin
  WNDs := EnumWndList(CPUWin.Handle);
  aHwnd := FindHwndFromList(WNDs, Str_TDisassemblerViewClass);
  DisComp := ObjectFromHWnd(aHwnd) as TWinControl;
  TMethod(FDisassemble) := RTTI_GetMethodAddr(DisComp, Str_DisasEvent);

  Old_P := RTTI_GetIntegerProp(DisComp, 'SelectedAddress');
  Old_TopP := RTTI_GetIntegerProp(DisComp, 'TopAddress');
  case StartType of
    stFromTopAddr : P := Old_TopP;
    stFromSelectAddr : P := Old_P;
  else
    P := StartType;
  end;

  while Line > 0 do
  begin
    RTTI_SetIntegerProp(DisComp, 'SelectedAddress', P);
    S_Source := RTTI_GetStringProp(DisComp, 'SelectedSource');

    //Get Next Address To P, and Get ASM Code to S_ASM
    FDisassemble(DisComp, P, S_ASM, L);

    if S_Source <> '' then
      Result := Result + S_Source + #$0D#$0A;
    Result := Result + S_ASM + #$0D#$0A;
    dec(Line);
  end;

  if Result <> '' then
    SetLength(Result, Length(Result) - 2);

  RTTI_SetIntegerProp(DisComp, 'TopAddress', Old_TopP);
  RTTI_SetIntegerProp(DisComp, 'SelectedAddress', Old_P);
end;

procedure TWizardMod.InstallExpert(Sender : TObject);
var
  Parent : HWND;
  Menus  : TPopupMenu;
begin
  if assigned(FOldExecute) then
  begin
    FOldExecute(Sender);
    RTTI_SetMethodAddr(DebugCPUCommand, Str_ExecuteCpuWin, TMethod(FOldExecute));
    FOldExecute := nil;
  end;

  Parent := HwndByClassName(Str_DisassemblyViewClass);
  if Parent = 0 then
    Exit;

  //find CPU Windows and Insert Menu
  CPUWin := ObjectFromHWnd(Parent) as TWinControl;
  if not Assigned(CPUWin) then
    Exit;

  //Insert Menus
  TObject(Menus) := CPUWin.FindComponent(Str_CPUPaneMenu);
  if not assigned(Menus) then
    Exit;

  with TCnMenuHook.Create(nil) do
  begin
    HookMenu(Menus);
    AddUserMenuItem(TCnSepMenuItem.Create(ipLast, 'itemSpec'));
    AddUserMenuItem(TCnUserMenuItem.Create('itemCopy50Line', 'Copy Special Lines', Copy50Line, ipLast, 'itemCopy50Line'));
    AddUserMenuItem(TCnUserMenuItem.Create('itemCopyLine', 'Copy Line...', CopyLine, ipLast, 'itemCopyLine'));
  end;
end;

constructor TWizardMod.Create;
var
  pMethodAddr : Pointer;
  Parent : HWND;
  FNewExecute : TNotifyEvent;
begin
  inherited;

  FOldExecute := nil;
  FNewExecute := InstallExpert;

  // 1. find Action Application.AppBulider.DebugCPUCommand
  Parent := HwndByClassName(Str_Delphi32AppClass);
  MainForm := ObjectFromHWnd(Parent);
  if not assigned(MainForm) then
    Exit;

  DebugCPUCommand := MainForm.FindComponent(Str_CpuCommandAction);
  if not assigned(DebugCpuCommand) then
    Exit;

  // 2. Hook Install Procedure to OnExecute Event
  TMethod(FOldExecute) := RTTI_GetMethodAddr(DebugCPUCommand, Str_ExecuteCpuWin);
  RTTI_SetMethodAddr(DebugCPUCommand, Str_ExecuteCpuWin, TMethod(FNewExecute));
end;

end.
